# -*- coding: utf-8 -*-

import os, logging, json, time, sys
from functools import partial

from qgis.gui import *
from qgis.core import *
from PyQt5 import QtGui, QtWidgets, uic
from PyQt5.QtCore import *
from qgis.utils import iface

from .rdi_tools import clearLayout
from .rdi_meta import rdiMeta

class rdiMap():
    def __init__(self, parent):
        self.dock = parent
        self.meta = rdiMeta()
        self.layer = None
        self.slider = None
        
    def paint(self, layout):
        if not self.dock.param.get('backMap'):
            return
        l = QtWidgets.QLabel()
        layout.addWidget(l)
        gb = QtWidgets.QGroupBox("Fond de plan")
        layout.addWidget(gb)
        lgb = QtWidgets.QVBoxLayout(gb) 
        self.layout = lgb
        backMap = self.meta.readIni("backMap")
        if isinstance(backMap, list) and len(backMap)>2:
            self.showChoice(backMap)
        else:
            self.askChoice()
    
    def askChoice(self):
        lgb = self.layout
        clearLayout(lgb)
        combo = QgsMapLayerComboBox()
        combo.setFilters(QgsMapLayerProxyModel.RasterLayer)
        lgb.addWidget(combo)
        butt = QtWidgets.QPushButton("Mémoriser")
        butt.setFixedSize(130, 30)
        lgb.addWidget(butt)
        lgb.setAlignment(butt, Qt.AlignCenter)
        butt.clicked.connect(partial(self.setBackMap, combo))
    
    def setLabel(self, wOri, wDest):
        op = wOri.value()
        wDest.setText(f"{op}%")
        self.opacity = op
        self.saveOpacity()
     
    def showChoice(self, backMap):       
        lgb = self.layout
        clearLayout(lgb)
        
        hbox = QtWidgets.QHBoxLayout()
        lgb.addLayout(hbox)
        
        name = backMap[2]
        if len(name)>40:
            name = name[:10] +"[...]"+ name[-25:]
        l = QtWidgets.QLabel(f"[{backMap[1]}] {backMap[0]}\n{name}")
        l.setToolTip(backMap[2])
        l.setWordWrap(True)
        font = QtGui.QFont()
        font.setItalic(True)
        l.setFont(font)
        hbox.addWidget(l)
        
        ico = QtGui.QIcon(os.path.join(self.dock.imageFolder,'refresh.png'))
        butt = QtWidgets.QPushButton(ico, '')
        butt.setToolTip("Actualiser")
        butt.setMaximumWidth(20)
        hbox.addWidget(butt)
        butt.clicked.connect(self.affMap)
        
        hbox = QtWidgets.QHBoxLayout()
        lgb.addLayout(hbox)
        l = QtWidgets.QLabel("Opacité")
        hbox.addWidget(l)
        w = QtWidgets.QSlider(Qt.Horizontal)
        w.setRange(0, 100)
        hbox.addWidget(w)
        l = QtWidgets.QLabel()
        hbox.addWidget(l)
        w.valueChanged.connect(partial(self.setLabel, w, l))  
        op = self.getOpacity(backMap)
        w.setValue(op)
        w.sliderReleased.connect(self.saveOpacity)
        self.slider = w
        
        butt = QtWidgets.QPushButton("Supprimer")
        butt.setFixedSize(130, 30)
        lgb.addWidget(butt)
        lgb.setAlignment(butt, Qt.AlignCenter)
        butt.clicked.connect(self.deleteBackMap)
    
    def deleteBackMap(self):
        self.meta.writeIni("backMap", "")
        self.askChoice()
        self.affMap()
    
    def setBackMap(self, combo):
        layer = combo.currentLayer()
        provider = layer.dataProvider()
        uri = provider.dataSourceUri()
        providerType = provider.name()
        layername = layer.name()
        backMap = [layername, providerType, uri, 100]
        self.meta.writeIni("backMap", backMap)
        self.showChoice(backMap)
        self.affMap()
    
    def getOpacity(self, backMap):
        try:
            op = int(backMap[3])
        except:
            op = 100
        return op
        
    def setOpacity(self, op):
        try:
            self.layer.layer().setOpacity(op/100)
            self.layer.layer().triggerRepaint()
        except:
            try:
                self.layer.layer().renderer().setOpacity(op/100)
                self.layer.layer().triggerRepaint()
            except:
                pass
     
    def actuSlider(self):
        try:
            try:
                opacity = self.layer.layer().opacity()
            except:
                opacity = self.layer.layer().renderer().opacity()
            op = int(opacity*100)
            self.slider.setValue(op)
        except:
            pass
    
    def saveOpacity(self):
        self.setOpacity(self.opacity)
        backMap = self.meta.readIni("backMap")
        backMap[3] = self.opacity
        self.meta.writeIni("backMap", backMap)
    
    def affMap(self):
        backMap = self.meta.readIni("backMap")
        if not self.dock.param.get('backMap') or not isinstance(backMap, list) or len(backMap)<3:
            self.layer = self.dock.dyn.removeMapLayer(self)
            return
        uri = backMap[2]
        providerType = backMap[1]
        layername = backMap[0]
        op = self.getOpacity(backMap)
        self.vlayer = QgsRasterLayer(uri, layername, providerType)
        self.layer = self.dock.dyn.addInGroup(self, 'background')
        self.setOpacity(op)
        self.layerId = self.layer.layerId()
        self.vlayer.repaintRequested.connect(self.actuSlider)
        self.dock.zoomIni.go()



class rdiZoom():
    def __init__(self, parent):
        self.dock = parent
        self.meta = rdiMeta()
        self.layer = None
        self.slider = None
        self.timer = QTimer()
        self.timer.setSingleShot(True)
        self.timer.timeout.connect(self.zoom)
        
    def paint(self, layout):
        l = QtWidgets.QLabel()
        layout.addWidget(l)
        gb = QtWidgets.QGroupBox("Zoom initial")
        layout.addWidget(gb)
        lgb = QtWidgets.QVBoxLayout(gb) 
        self.layout = lgb
        zoomIni = self.meta.readIni("zoomIni")
        if isinstance(zoomIni, list) and len(zoomIni)>3:
            self.showChoice(zoomIni)
        else:
            self.askChoice()
            
    def showChoice(self, zoomIni):
        lgb = self.layout
        clearLayout(lgb)
        
        w = QtWidgets.QWidget()
        lgb.addWidget(w)
        hbox = QtWidgets.QHBoxLayout(w)
        hbox.setSpacing(0)
        hbox.setContentsMargins(0,0,0,0)
        
        l = QtWidgets.QLabel(f"[{zoomIni[0]}   {zoomIni[1]}   {zoomIni[2]}   {zoomIni[3]}]")
        hbox.addWidget(l)
        
        ico = QtGui.QIcon(os.path.join(self.dock.imageFolder,'zoom.png'))
        butt = QtWidgets.QPushButton(ico, '')
        butt.setToolTip("Actualiser")
        butt.setMaximumWidth(20)
        hbox.addWidget(butt)
        butt.clicked.connect(self.zoom)
        
        butt = QtWidgets.QPushButton("Supprimer")
        butt.setFixedSize(130, 30)
        lgb.addWidget(butt)
        lgb.setAlignment(butt, Qt.AlignCenter)
        butt.clicked.connect(self.deleteZoomIni)
        
    def go(self):
        self.timer.start(1000)
    
    def zoom(self):
        if len(self.dock.alea.getEltsSelected())==0 and self.dock.mask.table is None:
            box = self.meta.readIni("zoomIni")
            if isinstance(box, list) and len(box)>3:
                rect = QgsRectangle(QRectF(QPointF(float(box[0]), float(box[1])), QPointF(float(box[2]), float(box[3]))))
                self.dock.zoomExtent(rect)
            else:
                self.dock.askZoom()
    
    def askChoice(self):
        lgb = self.layout
        clearLayout(lgb)
        l = QtWidgets.QLabel("Zoom automatique sur l'emprise des couches")
        lgb.addWidget(l)
        font = QtGui.QFont()
        font.setItalic(True)
        l.setFont(font)
        butt = QtWidgets.QPushButton("Emprise actuelle")
        butt.setFixedSize(130, 30)
        lgb.addWidget(butt)
        lgb.setAlignment(butt, Qt.AlignCenter)
        butt.clicked.connect(self.setZoomIni)
        
    def setZoomIni(self):
        rect = iface.mapCanvas().extent()
        tl = rect.toRectF().topLeft()
        br = rect.toRectF().bottomRight()
        coords = [round(tl.x()), round(tl.y()), round(br.x()), round(br.y())]
        # coords = [tl.x(), tl.y(), br.x(), br.y()]
        self.meta.writeIni("zoomIni", coords)
        self.showChoice(coords)
        
    def deleteZoomIni(self):
        self.meta.writeIni("zoomIni", "")
        self.askChoice()